;Inker.asm: firmware to drive a HP model 40 inkjet cartridge.
;This software is licensed under the GNU GPL.
;(c) 2006 Jeroen Domburg

;If you want to adjust the text this device prints, look at the bottom
;of this file.

;We're using a tiny2313 here.
.device ATTiny2313
;You might want to include the ATTiny2313-definitions included with
;your assembler. I use gavrasm, which doesn't need them.
;.include "tn2313def.inc"

;The fuses should be configured as follows:
;FUSE_L=0xc4
;FUSE_H=0xdf

;Timing variables. Play with this to maximize the contrast of the output

;The duration any nozzle is activated. Pick this too short and some nozzles
;won't shoot. Pick this too long and the DC/DC-converter won't be able to
;keep up.
.equ TIMESHOOT		= 15
;Cooling time. Longer = lighter output, shorter = more power usage, which again
;can mean the DC-DC-converter can't keep up.
.equ TIMECOOL		= 30
;Dots per pixel. Higher number = text takes longer to be completely written. If
;you adjust the speed you pull the nozzle over the paper accordingly, the output
;will become more dense.
.equ DOTSPERPIXEL	= 12
;If this value is 1, the pwm generator will be turned off when printing a space.
;If this value is 0, the pwm generator will continue running. Make this 0 when
;there are large number of spaces next to eachother in the text.
.equ PWMOFFWHENSPACE	= 0

.def dat	=	r0	;Data from an lpm instruction ends up here
.def temp	=	R16 	;The temp register
.def temp2	=	R17 	;Second temp register
.def temp3	=	R18
.def row1	=	r21
.def row2	=	r22
.def row3	=	r23
.def row4	=	r24
.def textpos	=	r25
.def charpos	=	r26
.def zlo	=	r30
.def zhi	=	r31


	rjmp start		; reset vector
	reti			; External Int 0 not enabled
	reti			; External Int 1 not enabled
	reti			; Timer 1 Capture Int not enabled
	reti			; Timer 1 Compare match A Int not enabled
	reti			; Timer 1 Overflow Int not enabled
	reti			; Timer 0 Overflow Int not enabled
	reti			; UART ReceiveInt
	reti			; UART Date Reg Empty Int not enabled
	reti			; UART tx'ed Int Not enabled
	reti			; Anacomp interrupt Not enabled


start:
	ldi temp,RAMEND		; top of memory
	out SPL,temp		; init stack pointer
	ldi temp,$ff		; Port B direction options
	out DDRB,temp		; Setup port direction
	ldi temp,0b100110	; Port D direction options
	out DDRD,temp		; Setup port direction
	ldi temp,$11		; Port A direction options
	out DDRA,temp		; Setup port direction
	
	
	;Set up timer 0 as a pwm generator to drive the DC-DC-converter
	ldi temp,0b100011
	out tccr0a,temp
	ldi temp,0b01
	out tccr0b,temp
	ldi temp,$20
	out ocr0b,temp

vast:
	;Reset textpos and charpos so the next code will pick up the first char
	ldi textpos,0
	ldi charpos,9
loop:
	inc charpos	;Next column of the character
	cpi charpos,10	;Final row?
	brne nonewchar	;No -> skip the next part
	
	ldi charpos,0	;reset charpos

	;put the position of the next character in z
	ldi zlo,low(text*2)
	ldi zhi,high(text*2)
	ldi temp,0
	add zlo,textpos
	adc zhi,temp
	inc textpos	;we're at the next character now
	lpm		;get the ascii value
	cp temp,dat	;=0? (signals end of string)
	breq vast	;Yep -> restart from the beginning
	ldi temp2,32	;Character data starts from ascii 32
	sub dat,temp2

;This might not work with your assembler. Just comment out the if
;and endif if that's the case.
.if (PWMOFFWHENSPACE==1)
	;Re-enable pwm generator
	ldi temp2,0b01
	out tccr0b,temp2
	ldi temp2,0	;Char is space?
	cp dat,temp2
	brne notspace
	;Yep -> disable pwm-gen to save power and keep the voltage from 
	;rising too high
	ldi temp2,0b01000000
	out tccr0b,temp2
notspace:
.endif
	
	;Character row-data is at z+dat*8
	ldi zlo,low(font*2)
	ldi zhi,high(font*2)
	;dat*8
	lsl dat
	rol temp
	lsl dat
	rol temp
	lsl dat
	rol temp
	;z=z+dat*8
	add zlo,dat
	adc zhi,temp
nonewchar:
	;We'll need to get the next column of character data
	ldi temp,0
	cpi charpos,8	;charpos>8?
	brsh emptyline	;Yep -> we're in the inter-character-blank space.
    	lpm		;Get char column data
	mov temp,dat
	adiw zlo,1	;Select next column
emptyline:
	;Fill the row data according to the character. This just assigns
	;4 nozzles to each bit in the column data.
	ldi row1,0
	ldi row2,0	
	ldi row3,0	
	ldi row4,0	
	sbrc temp,0
	 ori row1,$0f
	sbrc temp,1
	 ori row1,$f0
	sbrc temp,2
	 ori row2,$0f
	sbrc temp,3
	 ori row2,$f0
	sbrc temp,4
	 ori row3,$0f
	sbrc temp,5
	 ori row3,$f0
	sbrc temp,6
	 ori row4,$0f
	sbrc temp,7
	 ori row4,$f0

	;Send the column data
	rcall sendmul
	rjmp loop


	;Send the nozzle data DOTSPERPIXEL times, so the output won't be so compressed
sendmul:
	ldi temp,DOTSPERPIXEL
sendmul1:
	push temp
	rcall sendthis
	pop temp
	dec temp
	brne sendmul1
	ret
	
	;Send the data in the row registers one time.
sendthis:
	ldi temp2,1	;Select 1st nozzle
sendnext:
	mov temp,row1	;Sector data
	and temp,temp2	;Mask the selected nozzle
	out portb,temp	;and enable it
	sbi portd,1	;Enable sector 1
	rcall waitnozzle ;wait for the drop to fire
	cbi portd,1	;disable sector 1
	rcall waitcool	;and wait for the nozzle to cool.

	;Same with sector 2 to 4.
	mov temp,row2
	and temp,temp2
	out portb,temp
	sbi porta,1
	rcall waitnozzle
	cbi porta,1
	rcall waitcool
	
	mov temp,row3
	and temp,temp2
	out portb,temp
	sbi porta,0
	rcall waitnozzle
	cbi porta,0
	rcall waitcool

	mov temp,row4
	and temp,temp2
	out portb,temp
	sbi portd,2
	rcall waitnozzle
	cbi portd,2
	rcall waitcool

	lsl temp2	;Next nozzle
	cpi temp2,0	;All nozzles done?
	brne sendnext	;No -> loop.

	;Check and adjust the pwm value according to the output of the dc-dc converter.
	in temp,ocr0b	;get current pwm value
	subi temp,2	;assume it's too high
	sbis pind,6	;>20V?
	 subi temp,-4	;Yep -> cancel the assumed value and add 2.
	cpi temp,$f0	;Underflow?
	brlo noudf	;Nope -> skip
	ldi temp,0	;Undo underflow
noudf:
	cpi temp,$50	;Overflow?
	brlo noovf	;Nope -> skip
	ldi temp,$50	;Undo overflow
noovf:
	out ocr0b,temp	;and set the pwm value

	ret	;end of the routine :)

;Loop to wait for the nozzle to cool
waitcool:
	push temp
	ldi temp,TIMECOOL
cwait:
	nop
	dec temp
	brne cwait
	pop temp
	ret

;Loop to wait while the nozzle is firing.
waitnozzle:
	push temp
	ldi temp,TIMESHOOT
nwait:
	dec temp
	brne nwait
	pop temp
	ret

;The text the unit sends. Don't forget the ",0 on the end.
text:
    .db "http://www.elektuur.nl/         ",0
	
;Include the font file.
font:
.include "font8x8nr.inc" ;8x8 font